package cmd

var license = "" +
	"\n" +
	"  The MIT License (MIT)\n" +
	"\n" +
	"  Copyright (c) 2018 Tom Payne\n" +
	"\n" +
	"  Permission is hereby granted, free of charge, to any person obtaining a copy\n" +
	"  of this software and associated documentation files (the \"Software\"), to\n" +
	"  deal in the Software without restriction, including without limitation the\n" +
	"  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or\n" +
	"  sell copies of the Software, and to permit persons to whom the Software is\n" +
	"  furnished to do so, subject to the following conditions:\n" +
	"\n" +
	"  The above copyright notice and this permission notice shall be included in\n" +
	"  all copies or substantial portions of the Software.\n" +
	"\n" +
	"  THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n" +
	"  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n" +
	"  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\n" +
	"  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\n" +
	"  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING\n" +
	"  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS\n" +
	"  IN THE SOFTWARE.\n" +
	"\n" +
	""
